
from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])

p = instrument.new_program('baseband')
p.repetitions = 3

P1 = p.P1
P2 = p.P2

p.R.amplitude = 0.125
P1.Rs.amplitude = -0.125

P1.block_pulse(200, p.R.amplitude)
P1.block_pulse(200, P1.Rs.amplitude)
P1.block_pulse(200, p.R.amplitude + 0.5)
p.wait(1000)
P1.block_pulse(100, 0.5)
P2.block_pulse(100, -0.5)

p.wait(1000)
with p.parallel():
    P1.block_pulse(100, 0.25)
    P2.block_pulse(100, -0.25)

p.add_comment('--- long wait')
p.wait(100_000)
P1.block_pulse(500, -1.0)
p.add_comment('--- longer wait')
p.wait(160_000)
P1.block_pulse(500, 0.5)
p.add_comment('--- very long wait')
p.wait(500_000)
P1.block_pulse(1000, 0.75)
P2.block_pulse(1000, -0.75)
#
#p.describe()
#print()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



import os
import time
import logging
from contextlib import contextmanager
from numbers import Number

from .lang.conditions import CounterFlags
from .lang.exceptions import Q1InternalError, Q1ValueError
from .lang.triggers import TriggerCounter, Trigger
from .lang.math_expressions import Expression
from .lang.timeline import Timeline
from .lang.registers import Registers
from .lang.register import Register
from .lang.register_statements import RegisterAssignment
from .lang.loops import RangeLoop, LinspaceLoop, ArrayLoop
from .assembler.generator import Q1asmGenerator

logger = logging.getLogger(__name__)


class Program:
    verbose = False

    def __init__(self, path=None):
        self.sequence_builders = {}
        self.path = path if path is not None else os.path.join("q1", "_prog")
        self.R = Registers(self, local=False)
        self.repetitions = 1
        self._q1asm = {}
        self._loop_cnt = 0
        self._triggers = []
        # shared timeline for all sequencers
        self._timeline = Timeline()

    def add_sequence_builder(self, sequence_builder):
        name = sequence_builder.name
        self.sequence_builders[name] = sequence_builder
        setattr(self, name, sequence_builder)
        sequence_builder.start_sequence(self, self._timeline)

    def __getitem__(self, item):
        if item not in self.sequence_builders:
            raise Exception(f"no sequencer named {item}")
        return self.sequence_builders[item]

    def describe(self):
        for builder in self.sequence_builders.values():
            builder.describe()

    def seq_filename(self, name):
        os.makedirs(self.path, exist_ok=True)
        return os.path.join(self.path, f"q1seq_{name}.json")

    def compile(self, annotate=False, add_comments=True,
                listing=False, json=True, optimize=1):
        # store compiled sequences
        self._q1asm = {}

        start_compile = time.perf_counter()
        for builder in self.sequence_builders.values():
            g = Q1asmGenerator(add_comments=add_comments,
                               optimize=optimize)
            g.repetitions = self.repetitions
            start = time.perf_counter()
            builder.compile(g, annotate=annotate)
            end = time.perf_counter()
            d1 = (end-start)*1000
            start = end
            filename = self.seq_filename(builder.name) if listing or json else None
            g.assemble(listing=listing, json_output=json, filename=filename)
            self._q1asm[builder.name] = g.q1asm
            end = time.perf_counter()
            d2 = (end-start)*1000
            if Program.verbose:
                logger.debug(f"compile {builder.name} {d1:5.2f} {d2:5.2f} ms")
        duration = time.perf_counter() - start_compile
        logger.debug(f"Total compilation {duration*1000:5.2f} ms")

    def q1asm(self, name):
        return self._q1asm[name]

    def _add_statement(self, statement, init_section=False):
        if not isinstance(statement, RegisterAssignment):
            raise Q1InternalError(f"Illegal statement for program {statement}")
        for builder in self.sequence_builders.values():
            builder._add_statement(statement, init_section=init_section)

    def loop_range(self, start_stop, stop=None, step=None):
        ''' range loop '''
        return self.__loop(RangeLoop(self._loop_cnt, start_stop, stop, step))

    def loop_linspace(self, start, end, n, endpoint=True):
        ''' repeat loop '''
        return self.__loop(LinspaceLoop(self._loop_cnt, start, end, n, endpoint))

    def loop_array(self, values):
        ''' array loop '''
        return self.__loop(ArrayLoop(self._loop_cnt, values))

    @contextmanager
    def __loop(self, loop):
        self._loop_cnt += 1
        for s in self.sequence_builders.values():
            s.enter_loop(loop)

        yield loop.loopvar

        for s in self.sequence_builders.values():
            s.exit_loop(loop)

    @contextmanager
    def parallel(self):
        self._timeline.disable_update()
        yield
        self._timeline.enable_update()

    @contextmanager
    def conditional(self, counters, t_offset=0, evaluation_time=0):
        for s in self.sequence_builders.values():
            s.enter_conditional(counters, t_offset=t_offset, evaluation_time=evaluation_time)
        flags = CounterFlags(self)
        yield flags
        for s in self.sequence_builders.values():
            s.exit_conditional()

    @contextmanager
    def condition(self, operator):
        for s in self.sequence_builders.values():
            s.enter_condition(operator)
        yield
        for s in self.sequence_builders.values():
            s.exit_condition()

    def configure_trigger(self, sequencer_name, invert=False):
        addr = len(self._triggers)+1
        trigger = Trigger(sequencer_name, address=addr, invert=invert)
        self._triggers.append(trigger)
        self.sequence_builders[sequencer_name].trigger = trigger
        return trigger

    def add_trigger_counter(self, trigger, threshold=1, invert=False):
        counter = TriggerCounter(trigger, threshold=threshold, invert=invert)
        for s in self.sequence_builders.values():
            s._add_trigger_counter(counter)
        return counter

    def latch_enable(self, enable, t_offset=0):
        for s in self.sequence_builders.values():
            s.latch_enable(enable, t_offset=t_offset)

    def latch_reset(self, t_offset=0):
        for s in self.sequence_builders.values():
            s.latch_reset(t_offset=t_offset)

    def add_comment(self, comment):
        for s in self.sequence_builders.values():
            s.add_comment(comment)

    def wait(self, t):
        if isinstance(t, Number):
            if t < 0:
                raise Q1ValueError("Wait time must be positive")
            self._timeline.set_pulse_end(self._timeline.current_time + t)
        else:
            for s in self.sequence_builders.values():
                s._add_reg_wait(t)

    def block_pulse(self, duration, sequencers, amplitudes, t_offset=0):
        if not isinstance(duration, (Register, Expression)):
            self._timeline.disable_update()
            for s, a in zip(sequencers, amplitudes):
                if isinstance(s, str):
                    s = self[s]
                s.block_pulse(duration, a, t_offset=t_offset)
            self._timeline.enable_update()
        else:
            if not self._timeline.is_running:
                raise Q1ValueError("Variable pulse length not possible in parallel section")
            self.set_offsets(sequencers, amplitudes, t_offset=t_offset)
            self.wait(duration)
            self.set_offsets(sequencers, [0.0]*len(sequencers), t_offset=0)

    def ramp(self, duration, sequencers, v_start, v_stop, t_offset=0):
        self._timeline.disable_update()
        for s, v1, v2 in zip(sequencers, v_start, v_stop):
            if isinstance(s, str):
                s = self[s]
            s.ramp(duration, v1, v2, t_offset=t_offset)
        self._timeline.enable_update()

    def set_offsets(self, sequencers, offsets, t_offset=0):
        self._timeline.disable_update()
        for s, offset in zip(sequencers, offsets):
            if isinstance(s, str):
                s = self[s]
            s.set_offset(offset, t_offset=t_offset)
        self._timeline.enable_update()



import numpy as np
import matplotlib.pyplot as pt

from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('q1', qcm0.name, [0, 1], nco_frequency=200e6)
instrument.add_readout('rf1', qrm1.name, [0, 1], nco_frequency=0e6)

qrm1.in0_gain(0)
qrm1.in1_gain(0)
# %%
amplitude = 0.5
N = 101
period = 1000
f_start = -250e6
f_stop = 250e6

p = instrument.new_program('acq_rf_sweep')
p.repetitions = 1

q1 = p.q1
rf1 = p.rf1

rf1.add_acquisition_bins("default", N)
rf1.integration_length_acq = 1000
rf1.nco_prop_delay = 148  # ~ 50 cm line

q1.set_offset(amplitude, 0.0)
rf1.set_offset(amplitude)
rf1.acquire_frequency_sweep(N, period,
                            f_start, f_stop,
                            "default",
                            acq_delay=152)
rf1.wait(period)

rf1.set_offset(0.0)
q1.set_offset(0.0)

p.compile(listing=True)

# %%
instrument.run_program(p)


data = instrument.get_acquisition_bins('rf1', 'default')

# pprint(data)
I = np.array(data['integration']['path0'])/rf1.integration_length_acq
Q = np.array(data['integration']['path1'])/rf1.integration_length_acq
c = I + 1j*Q
# for i in range(N):
#     print(f'{I[i]:6.3f}, {Q[i]:6.3f}, {abs(c[i]):6.3f}, {np.angle(c[i])/np.pi:5.3f}')

pt.close('all')
pt.figure()
pt.plot(np.angle(c)/np.pi)
pt.title('phase')

pt.figure()
pt.plot(np.abs(c))
pt.title('amplitude')
pt.ylim(0, 0.6)

np.abs(c)



import init_logger
from q1pulse.instrument import Q1Instrument
import numpy as np

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('q1', qcm0.name, [0,1], nco_frequency=100e6)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_readout('R1', qrm1.name, [])

p = instrument.new_program('shaped_block')
p.repetitions = 2

q1 = p.q1
P1 = p.P1
P2 = p.P2
R1 = p.R1


block80 = q1.add_wave('block80', np.ones(80))
block80 = P1.add_wave('block80', np.ones(80))

R1.add_acquisition_bins('acq_bins', p.repetitions)
R1.integration_length_acq = 80

amplitude = 0.125
p.R.wait = 20

with p.loop_range(4):
    q1.shaped_pulse(block80, 0.5, block80, 0.5)
    P1.shaped_pulse(block80, 0.1)
    p.wait(20)
    p.wait(p.R.wait)

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0, qrm1])

if hasattr(qrm1, 'print_acquisitions'):
    # get acquisition timing and counts from Q1Simulator
    qrm1.print_acquisitions()



from __future__ import annotations

import logging
import signal
import threading
from collections.abc import Callable
from typing import Any, cast

from types import FrameType, TracebackType

logger = logging.getLogger(__name__)


class DelayedKeyboardInterrupt:
    signal_received: tuple[int, FrameType | None] | None = None
    # the handler type is seemingly only defined in typesheeed so copy it here
    # manually https://github.com/python/typeshed/blob/main/stdlib/signal.pyi
    old_handler: Callable[[int, FrameType | None], Any] | int | None = None

    def __init__(self, where: str = "") -> None:
        """
        A context manager to wrap a piece of code to ensure that a
        KeyboardInterrupt is not triggered by a SIGINT during the execution of
        this context. A second SIGINT will trigger the KeyboardInterrupt
        immediately.

        Inspired by https://stackoverflow.com/questions/842557/how-to-prevent-a-block-of-code-from-being-interrupted-by-keyboardinterrupt-in-py
        """
        self._where_msg = f"Section: {where}" if where != "" else ""

    def __enter__(self) -> None:
        is_main_thread = threading.current_thread() is threading.main_thread()
        if not is_main_thread:
            # logger.debug(
            #     "Not on main thread. Cannot intercept interrupts. " + self._where_msg
            # )
            return
        handler = signal.getsignal(signal.SIGINT)
        try:
            is_delayed_sig_handler = handler.__self__.__class__ == DelayedKeyboardInterrupt
        except Exception:
            is_delayed_sig_handler = False
        if not is_delayed_sig_handler:
            self.old_handler = signal.signal(signal.SIGINT, self._handler)

    def _handler(self, sig: int, frame: FrameType | None) -> None:

        self.signal_received = (sig, frame)
        print(
            "Received SIGINT, Will interrupt at first suitable time. "
            "Send second SIGINT to interrupt immediately."
        )
        # now that we have gotten one SIGINT make the signal
        # trigger a keyboard interrupt normally
        signal.signal(signal.SIGINT, self._forceful_handler)
        logger.info("SIGINT received. Delaying KeyboardInterrupt." + self._where_msg)

    def _forceful_handler(self, sig: int, frame: FrameType | None) -> None:
        print("Second SIGINT received. Triggering KeyboardInterrupt immediately.")
        logger.info(
            "Second SIGINT received. Triggering KeyboardInterrupt immediately. "
        )
        # The typing of signals seems to be inconsistent
        # since handlers must be types to take an optional frame
        # but default_int_handler does not take None.
        # see https://github.com/python/typeshed/pull/6599
        frame = cast("FrameType", frame)
        signal.default_int_handler(sig, frame)

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if self.old_handler is not None:
            signal.signal(signal.SIGINT, self.old_handler)
        if self.signal_received is not None:
            if self.old_handler is not None and not isinstance(self.old_handler, int):
                self.old_handler(*self.signal_received)



import json
import os
import time
import logging
from collections import defaultdict
from pathlib import Path
from tempfile import TemporaryDirectory


from q1pulse.program import Program
from q1pulse.lang.exceptions import Q1InputOverloaded, Q1InternalError
from q1pulse.sequencer.sequencer import SequenceBuilder
from q1pulse.sequencer.control import ControlBuilder
from q1pulse.sequencer.readout import ReadoutBuilder
from q1pulse.turbo_cluster import TurboCluster
from q1pulse.modules.modules import QcmModule, QrmModule, QbloxModule, Sequencer
from q1pulse.modules.sequencer_states import translate_seq_status
from q1pulse.util.delayedkeyboardinterrupt import DelayedKeyboardInterrupt
from q1pulse.util.qblox_version import check_qblox_instrument_version

logger = logging.getLogger(__name__)


class Q1Instrument:
    verbose = False
    concurrent_communication = True
    ignore_acq_binning_done = False

    # Postpone error checking till the end to save communication overhead.
    # System errors are only reported for SCPI errors. It's higly unlikely to
    # get an error, because everything is already checked in qblox-instruments code.
    # Note on v0.16: Default cluster behavior is like _i_feel_lucky = True.
    _i_feel_lucky = True

    _exception_on_overload = True

    def __init__(self, path=None, add_traceback=True):
        check_qblox_instrument_version()
        if path:
            self.path = path
        else:
            q1dir = Path.home() / ".q1"
            q1dir.mkdir(exist_ok=True)
            self.temp_dir = TemporaryDirectory(dir=q1dir)
            self.path = self.temp_dir.name
            logger.info("Instrument upload temp dir: " + self.path)
        self.root_instruments = set()
        self.modules: dict[str, QbloxModule] = {}
        self.controllers: dict[int, Sequencer] = {}
        self.readouts: dict[int, Sequencer] = {}
        self._loaded_q1asm: dict[str, dict] = {}
        SequenceBuilder.add_traceback_to_instructions = add_traceback

    def add_qcm(self, module):
        logger.info(f"Add {module.name}")
        self.modules[module.name] = QcmModule(module)
        self._add_root_instrument(module.root_instrument)

    def add_qrm(self, module):
        logger.info(f"Add {module.name}")
        self.modules[module.name] = QrmModule(module)
        self._add_root_instrument(module.root_instrument)

    def add_control(self, name, module_name, channels, nco_frequency=None):
        sequencer = self.modules[module_name].get_sequencer(channels)
        sequencer.nco_frequency = nco_frequency
        sequencer.label = name
        self.controllers[name] = sequencer

    def add_readout(self, name, module_name, out_channels=[],
                    nco_frequency=None, in_channels=[0, 1]):
        module = self.modules[module_name]
        if not isinstance(module, QrmModule):
            raise Exception(f"Module {module_name} is not a QRM")
        sequencer = module.get_sequencer(out_channels)
        sequencer.nco_frequency = nco_frequency
        sequencer.in_channels = in_channels
        sequencer.label = name
        self.readouts[name] = sequencer

    def _add_root_instrument(self, root_instrument):
        self.root_instruments.add(root_instrument)

    def new_program(self, prog_name):
        program = Program(path=os.path.join(self.path, prog_name))
        for name, seq in self.controllers.items():
            seq_builder = ControlBuilder(name, seq.enabled_paths,
                                         seq.max_output_voltage,
                                         seq.nco_frequency)
            program.add_sequence_builder(seq_builder)

        for name, seq in self.readouts.items():
            seq_builder = ReadoutBuilder(name, seq.enabled_paths,
                                         seq.max_output_voltage,
                                         seq.nco_frequency)
            program.add_sequence_builder(seq_builder)

        return program

    def save_program(self, program: Program, path: str):
        """Stores program including current settings of the sequence builders,
        like nco_frequency.

        Args:
            program: program to save.
            path: path to store the program.
        """
        os.makedirs(path, exist_ok=True)
        config = {}
        for builder in program.sequence_builders.values():
            name = builder.name
            q1asm = program.q1asm(name)
            if q1asm is not None:
                filename = f"q1seq_{name}.json"
                with open(os.path.join(path, filename), "w", encoding="utf-8") as f:
                    json.dump(q1asm, f, indent=1, separators=(",", ":"))
            else:
                filename = None
            sequencer = self.controllers.get(name)
            if sequencer is not None:
                seq_type = "controller"
            else:
                sequencer = self.readouts.get(name)
                seq_type = "readout"
                if sequencer is None:
                    raise Q1InternalError(f"Sequencer {name} not found in instrument")
            builder_config = {
                "module": sequencer.module_name,
                "seq_nr": sequencer.seq_nr,
                "seq_type": seq_type,
                "out_channels": sequencer.channels,
                "nco": builder.nco_frequency,
                "paths": builder.enabled_paths,
                "sequence": filename,
                "duration": builder.end_time if q1asm is not None else None,
                }
            if seq_type == "readout":
                builder_config["in_channels"] = sequencer.in_channels
            config[name] = builder_config
        with open(os.path.join(path, "q1program.json"), "w", encoding="utf-8") as f:
            json.dump(config, f, indent=2)

    def run_program(self, program):
        self.start_program(program)
        self.wait_stopped()

    def start_program(self, program):
        t_start = time.perf_counter()

        for instrument in self.root_instruments:
            with DelayedKeyboardInterrupt("check status"):
                check_instrument_status(instrument)
                if Q1Instrument._i_feel_lucky and hasattr(instrument, "_debug"):
                    # Change the debug level to speed up communication.
                    # Errors will be checked before start of the sequence.
                    instrument._debug = 2

        instruments_with_sequence = set()
        sequencers = {**self.controllers, **self.readouts}
        n_configured = 0

        # sort on (seq_num, slot)
        sequencers = dict(sorted(sequencers.items(),
                                 key=lambda kv: (kv[1].seq_nr, self.modules[kv[1].module_name].slot_idx)))

        for name, seq in sequencers.items():
            t_start_seq = time.perf_counter()
            module = self.modules[seq.module_name]
            with DelayedKeyboardInterrupt("configure sequencers"):
                q1asm = program.q1asm(name)
                self._loaded_q1asm[name] = q1asm
                if q1asm is None:
                    module.disable_seq(seq)
                    module.set_awg_offsets(seq.seq_nr, 0.0, 0.0)
                    logger.debug(f"Sequencer {name} no sequence")
                    continue
                n_configured += 1
                instruments_with_sequence.add(module.root_instrument)
                module.set_label(seq.seq_nr, name)
                module.upload(seq.seq_nr, q1asm)
                module.invalidate_cache(seq.seq_nr, "offset_awg_path0")
                module.invalidate_cache(seq.seq_nr, "offset_awg_path1")
                module.enable_seq(seq)
                prog_seq = program[name]
                module.set_nco(seq.seq_nr, prog_seq.nco_frequency)
                if prog_seq.modifies_frequency:
                    module.invalidate_cache(seq.seq_nr, "nco_freq")
                if prog_seq.mixer_gain_ratio is not None:
                    module.set_mixer_gain_ratio(seq.seq_nr, prog_seq.mixer_gain_ratio)
                if prog_seq.mixer_phase_offset_degree is not None:
                    module.set_mixer_phase_offset_degree(seq.seq_nr, prog_seq.mixer_phase_offset_degree)
                # configure trigger counters
                for counter in prog_seq.trigger_counters:
                    module.configure_trigger_counter(seq.seq_nr, counter.trigger.address,
                                                     counter.threshold, counter.invert)
            if Q1Instrument.verbose:
                duration = time.perf_counter() - t_start_seq
                logger.debug(f"Configured {name} in {duration*1000.0:3.1f} ms")

        for name, seq in self.readouts.items():
            t_start_seq = time.perf_counter()
            module = self.modules[seq.module_name]
            if not module.enabled(seq.seq_nr):
                continue
            with DelayedKeyboardInterrupt("configure readout"):
                readout = program[name]
                module.thresholded_acq_rotation(seq.seq_nr, readout.thresholded_acq_rotation)
                module.thresholded_acq_threshold(seq.seq_nr, readout.thresholded_acq_threshold)
                module.integration_length_acq(seq.seq_nr, int(readout.integration_length_acq))
                module.nco_prop_delay(seq.seq_nr, int(readout.nco_prop_delay))
                module.delete_acquisition_data(seq.seq_nr)
                trigger = readout.trigger
                if trigger is not None:
                    module.set_trigger(seq.seq_nr, trigger.address, trigger.invert)
                else:
                    module.set_trigger(seq.seq_nr, None)
                module.set_ttl(
                    seq.seq_nr,
                    readout.ttl_acq_input_select,
                    readout.ttl_acq_threshold,
                    readout.ttl_acq_auto_bin_incr_en
                    )

            if Q1Instrument.verbose:
                duration = time.perf_counter() - t_start_seq
                logger.debug(f"Configured QRM {name} in {duration*1000.0:3.1f} ms")

        with DelayedKeyboardInterrupt("arm and start"):
            t_start_arm = time.perf_counter()
            # Note: arm per sequencer. Arm on the cluster still gives red leds on the modules.
            for module in self.modules.values():
                module.arm_sequencers()
            if Q1Instrument.verbose:
                duration = time.perf_counter() - t_start_arm
                logger.debug(f"Armed {n_configured} sequencers in {duration*1000.0:3.1f} ms")

            # Error check implicitly waits for the module to process all previous commands.
            # Exclude CMM (slot=0)
            self.check_system_errors(exclude=[0])

            for module in self.modules.values():
                module.start_sequencers()
            # for instrument in instruments_with_sequence:
            #     # t = (time.perf_counter() - t_start) * 1000
            #     # logger.debug(f"Start  ({t:5.3f} ms)")
            #     instrument.start_sequencer()
            self.check_system_errors()
            self._t_start = time.perf_counter()

        t = (time.perf_counter() - t_start) * 1000
        logger.info(f"Duration upload/start: ({t:5.3f}ms)")

    def wait_stopped(self, timeout_minutes=1):
        try:
            # Wait for completion
            errors = {}
            msg_level = 0
            # list[tuple[module, sequencer]]
            active_sequencers: list[tuple[QbloxModule, SequenceBuilder]] = []
            sequencers = {**self.controllers, **self.readouts}
            for name, seq in sequencers.items():
                module = self.modules[seq.module_name]
                if not module.enabled(seq.seq_nr):
                    continue
                active_sequencers.append((module, seq))

            statuses = self._get_sequencer_status_multiple(active_sequencers, timeout_minutes)
            for (module, seq), status in zip(active_sequencers, statuses):
                # status = self._get_sequencer_status(module, seq.seq_nr, timeout_minutes)
                if "ACQ BINNING DONE" in status.debug_msgs:
                    module.mark_acq_ready(seq.seq_nr)
                logger.log(status.level,
                           f"Status {module.slot_idx}:{seq.seq_nr} ({module.pulsar.name}:{seq.seq_nr}): {status}")
                msg_level = max(msg_level, status.level)
                if status.status != "OKAY" or status.state != "STOPPED" or status.level >= logging.WARNING:
                    errors[name] = str(status)
                    # reset awg offsets in case of any error.
                    with DelayedKeyboardInterrupt("set offsets"):
                        module.set_awg_offsets(seq.seq_nr, 0.0, 0.0)
                if status.input_overloaded:
                    if Q1Instrument._exception_on_overload:
                        raise Q1InputOverloaded(
                                f"INPUT OVERLOAD on {name}."
                                "\nException can be suppressed with q1pulse.set_exception_on_overload(False)")
                    else:
                        print(f"WARNING: input overload on {name}")

            if msg_level == logging.ERROR:
                logger.error("*** Program errors ***")
                for name, state in errors.items():
                    logger.error(f"  {name}: {state}")
                raise Exception(f"Q1 failures (see logging):\n {errors}")
            duration = time.perf_counter() - self._t_start
            logger.debug(f"Ready after {duration*1000:.1f} ms")
        except Exception:
            logger.error("Exception", exc_info=True)
            raise
        finally:
            with DelayedKeyboardInterrupt("stop sequencers"):
                logger.debug("Stop sequencers")
                # for instrument in self.root_instruments:
                #     instrument.stop_sequencer()
                for module in self.modules.values():
                    module.stop_sequencers()

    def _get_sequencer_status(self, module, seq_nr, timeout_minutes):
        """Get sequencer status in a interrupt safe way.
        Only intercept the keyboard interrupt during communication,
        not when sleeping.
        """
        expiration_time = time.perf_counter() + timeout_minutes*60.0
        # timeout_poll_res = 0.001
        with DelayedKeyboardInterrupt("check status"):
            status = module.get_sequencer_status(seq_nr, 0.0)
        while (status.state == "RUNNING"
                or status.state == "Q1_STOPPED"
                or status.state == "ARMED"
               ) and time.perf_counter() < expiration_time:
            # time.sleep(timeout_poll_res)
            with DelayedKeyboardInterrupt("check status"):
                status = module.get_sequencer_status(seq_nr, 0.0)
        return status

    def _get_sequencer_status_multiple(
            self,
            active_sequencers: list[tuple[QbloxModule, SequenceBuilder]],
            timeout_minutes):
        """Get sequencer status in a interrupt safe way.
        Only intercept the keyboard interrupt during communication,
        not when sleeping.
        """
        not_ready = ["ARMED", "RUNNING", "Q1_STOPPED"]
        statuses = []
        expiration_time = time.perf_counter() + timeout_minutes*60.0
        # timeout_poll_res = 0.001
        if not Q1Instrument.concurrent_communication:
            for module, seq in active_sequencers:
                statuses.append(self._get_sequencer_status(module, seq.seq_nr, timeout_minutes))
        else:
            statuses = [None]*len(active_sequencers)
            # reverse lookup in statuses
            index: dict[tuple[TurboCluster, int, int], int] = {}
            turbo_sequencers: dict[TurboCluster, dict[QbloxModule, list[int]]] = {}
            # group by instrument. Keep index in list.
            # if not TurboCluster use normal. sequential method.
            for idx, (module, seq) in enumerate(active_sequencers):
                instrument = module.root_instrument
                if not isinstance(instrument, TurboCluster):
                    statuses[idx] = self._get_sequencer_status(module, seq.seq_nr, timeout_minutes)
                else:
                    module_sequencers = turbo_sequencers.setdefault(instrument, defaultdict(list))
                    module_sequencers[module].append(seq.seq_nr)
                    index[(instrument, module.slot_idx, seq.seq_nr)] = idx

            while any(status is None or status.state in not_ready for status in statuses):
                for instrument, mod_seqs in turbo_sequencers.items():
                    # collect sequencers that are not yet ready
                    sequencers = {}
                    for module, seqs in mod_seqs.items():
                        seq_nums = []
                        slot = module.slot_idx
                        for seq_num in seqs:
                            status = statuses[index[(instrument, slot, seq_num)]]
                            if status is None or status.state in not_ready:
                                seq_nums.append(seq_num)
                        sequencers[slot] = seq_nums

                    try:
                        res = None
                        with DelayedKeyboardInterrupt("check status"):
                            res = instrument.get_sequencer_status_multiple(sequencers)
                    except KeyboardInterrupt:
                        logger.info(f"Interrupted during get_sequencer_status_multiple. Statusses: {res}")
                        raise

                    for slot, seq_num, status in res:
                        status = translate_seq_status(status)
                        statuses[index[(instrument, slot, seq_num)]] = status
                if time.perf_counter() > expiration_time:
                    break
                # if any(status.state in not_ready for status in statuses):
                #     time.sleep(timeout_poll_res)

        return statuses

    def check_system_errors(self, exclude: list[int] = []):
        t_start_check = time.perf_counter()
        errors = []

        for instrument in self.root_instruments:
            if Q1Instrument.concurrent_communication and isinstance(instrument, TurboCluster):
                errors += instrument.get_system_errors(exclude)
            else:
                while instrument.get_num_system_error() != 0:
                    errors.append(instrument.get_system_error())

        if not Q1Instrument.concurrent_communication:
            for module in self.modules.values():
                while module.get_num_system_error() != 0:
                    errors.append(module.get_system_error())

        if len(errors) > 0:
            if Q1Instrument._i_feel_lucky:
                logger.error("You're not lucky. One of the previous calls failed...")
            msg = instrument.name + ":" + "\n".join(errors)
            logger.error(msg)
            raise RuntimeError(msg)

        if Q1Instrument.verbose:
            duration = time.perf_counter() - t_start_check
            logger.debug(f"Checked errors in {duration*1000.0:3.1f} ms")

    def get_acquisition_bins(self, sequencer_name, acq_name):
        acq_data = self._get_acquisitions(sequencer_name, acq_name)
        if acq_data is None:
            return None
        return acq_data["acquisition"]["bins"]

    def get_scope_data(self, sequencer_name: str, acq_name: str):
        acq_data = self._get_acquisitions(sequencer_name, acq_name)
        if acq_data is None:
            return None
        return acq_data["acquisition"]["scope"]

    def _get_acquisitions(self, sequencer_name, acq_name):
        seq = self.readouts[sequencer_name]
        q1asm = self._loaded_q1asm[sequencer_name]
        if q1asm is None or len(q1asm["acquisitions"]) == 0:
            logger.warning(f"No acquisitions for {sequencer_name}")
            return None
        module = self.modules[seq.module_name]
        if self.ignore_acq_binning_done:
            finished = module.get_acquisition_status(seq.seq_nr, 0)
            if not finished:
                logger.info("Acquisition not finished (according to QRM)")
        else:
            # check if module is ready.
            timeout = 0.1  # wait max 6 s.
            finished = module.get_acquisition_status(seq.seq_nr, timeout)
            if not finished:
                logger.error("Acquisition not finished (according to QRM)")
        with DelayedKeyboardInterrupt("get_acquisitions"):
            return module.get_acquisitions(seq.seq_nr, acq_name)

    def get_input_ranges(self, sequencer_name):
        """ Returns input range for both channels of sequencer.
        Value is in Vpp.
        """
        seq = self.readouts[sequencer_name]
        module = self.modules[seq.module_name]
        if module.pulsar.is_rf_type:
            att_db = module.pulsar.in0_att.cache()
            in_range = 1.0 * 10**(att_db/20)
            in_range = (in_range, in_range)
        else:
            in0_gain = module.pulsar.in0_gain.cache()
            in1_gain = module.pulsar.in1_gain.cache()
            in_range = tuple(
                    1.0 * 10**(-dB/20)
                    for dB in [in0_gain, in1_gain])
        return in_range


def set_exception_on_overload(enable: bool):
    """
    Setting set_exception_on_overload to False suppresses Exception
    when the QRM input is overloaded.
    """
    Q1Instrument._exception_on_overload = enable


def check_instrument_status(instrument, print_status=False):
    t = time.perf_counter()
    sys_state = instrument.get_system_status()
    if Q1Instrument.verbose:
        d = time.perf_counter() - t
        logger.debug(f"Status {sys_state.status} {d*1000:.1f} ms")

    if sys_state.status != "OKAY":
        if getattr(instrument, "is_dummy", False):
            print(f"Status (Dummy) {instrument.name}:", sys_state)
        elif instrument.get_idn()["serial_number"] == "whatever":
            # looks like dummy cluster
            print(f"Status (Dummy) {instrument.name}:", sys_state)
        else:
            raise Exception(f"{instrument.name} status not OKAY: {sys_state}")




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qrm1.name, [0])

p = instrument.new_program('markers')
p.repetitions = 1000

P1 = p.P1
P2 = p.P2

with p.parallel():
    P1.block_pulse(20, -0.5)
    P2.block_pulse(20, -0.5)
p.wait(8)
with p.parallel():
    P1.set_markers(0b0001)
    P2.set_markers(0b0001)
    P1.block_pulse(20, 0.5)
    P2.block_pulse(20, 0.5)
P1.set_markers(0b0000)
P2.set_markers(0b0000)
p.wait(50_000)

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0, qrm1])



import numpy as np
from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

# Do not specify path. This is not a Q1ASM reference test.
instrument = Q1Instrument()
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qcm0.name, [0])
instrument.add_control('P2', qcm0.name, [1])

p = instrument.new_program('sync_wave')
p.repetitions = 10_000_000

N = 10
P1 = p.P1
P2 = p.P2

ramped = False

if ramped:
    block = np.concatenate([
            np.linspace(0,-0.5, 2, endpoint=False),
            -0.5*np.ones(N-4),
            np.linspace(-0.5, 0.5, 4, endpoint=False),
            0.5*np.ones(N-4),
            np.linspace(0.5, 0, 2, endpoint=False),
            ])
else:
    block = np.concatenate([
            -0.5*np.ones(N),
            0.5*np.ones(N),
            ])
P1.add_wave('block', block)
P2.add_wave('block', block)

P1.set_gain(1.0)
P2.set_gain(1.0)
with p.loop_range(2):
    with p.parallel():
        P1.play('block')
        P2.play('block')
        p.wait(2*N)

p.wait(10000)
p.describe()

p.compile(listing=True)

instrument.run_program(p)

plot_output([qcm0, qrm1])



import sys
import os
from copy import copy
from contextlib import contextmanager
import traceback
import logging

from .builderbase import BuilderBase
from ..lang.triggers import TriggerCounter
from ..lang.exceptions import (
        Q1StateError, Q1Exception,
        Q1InternalError, Q1SequenceError,
        Q1TimingError, Q1SyntaxError,
        )
from ..lang.sequence import Sequence
from ..lang.loops import Loop
from ..lang.registers import Registers
from ..lang.timed_statements import WaitRegStatement, TimedStatement
from ..lang.flow_statements import (
        LoopDurationStatement,
        LoopStatement, EndLoopStatement,
        ArrayLoopStatement, EndArrayLoopStatement,
        )
from ..lang.loops import LinspaceLoop, RangeLoop, ArrayLoop
from ..lang.simulator_statements import LogStatement
from ..lang.conditions import (
        LatchEnableStatement, LatchResetStatement,
        BranchSequence, ConditionalBlockStatement,
        CounterFlags,
        )

logger = logging.getLogger(__name__)


class SequenceBuilder(BuilderBase):
    add_traceback_to_instructions = True
    '''
    Adding traceback is convenient to find a problem in your script.
    However, it makes the compiler pretty slow.
    Q1Instrument can suppress the traceback.
    '''

    MIN_DURATION = 4

    def __init__(self, name):
        self.name = name
        self._local_loop_cnt = 0
        self.Rs = Registers(self, local=True)
        self._sequence_stack = []
        self.sequence = None
        self._local_time = 0
        self._local_time_active = False
        self._compiled = False
        self._init_sequence = Sequence(None)
        self._last_timed_statement = None
        self._conditional_block = None
        self._in_condition = False
        self._trigger_counters = []

    def start_sequence(self, program, timeline):
        self._program = program
        self._timeline = timeline
        self._sequence_push(Sequence(self._timeline))

    def _sequence_push(self, sequence):
        self._sequence_stack.append(sequence)
        self.sequence = sequence

    def _sequence_pop(self):
        self._sequence_stack.pop()
        self.sequence = self._sequence_stack[-1]

    def _add_statement(self, statement, init_section=False):
        self._check_time(statement)
        if SequenceBuilder.add_traceback_to_instructions and not isinstance(statement, str):
            self._add_traceback(statement)
        if self._compiled:
            raise Q1StateError('Program cannot be changed after compilation')
        if init_section:
            self._init_sequence.add(statement)
        else:
            self.sequence.add(statement)

    def _check_time(self, statement):
        # TODO: Distinguish parameter operations vs RT IO / RT control instructions.
        if not isinstance(statement, TimedStatement):
            return
        t = statement.time
        last = self._last_timed_statement
        if last and t != last.time:
            delta_t = t - last.time
            if delta_t < 0:
                raise Q1TimingError(f'Statement time cannot be before {last.time}\n'
                                    f'Previous: t={last.time}  {last}\n'
                                    f'Adding:   t={statement.time}  {statement}')
            if delta_t < SequenceBuilder.MIN_DURATION:
                raise Q1TimingError(f'Time between statements must be at least of 4 ns:\n'
                                    f'Adding:  t={statement.time}  {statement} (delta_t {delta_t})')
        self._last_timed_statement = statement

    def _add_traceback(self, statement):
        max_depth = 10
        # add 2 levels: _add_statement and _add_traceback.
        # these 2 levels are not added to statement.tb
        stack = traceback.extract_stack(limit=max_depth+2)
        tb = []
        tb.append('\nTraceback to Q1Pulse:\n')
        i = 0
        start = 0
        for entry in stack:
            i += 1
            # '<module>' is the root script (at least when running in Spyder)
            if entry.name == '<module>':
                start = i
        i = 0
        for entry in stack:
            i += 1
            if i < start:
                continue
            if i > max_depth:
                break
            tb.append(f'  File "{entry.filename}", line {entry.lineno}')
            tb.append(f'    {entry.line}\n')
        statement.tb = tb

    def wait(self, t):
        self.set_pulse_end(self.current_time + t)

    def add_comment(self, comment):
        self._add_statement(comment)

    def log(self, msg, var=None, time=False):
        '''
        Writes a value to the logging of Q1Simulator.
        It uses a special comment line in Q1ASM that is ignored by
        the Pulsar.
        '''
        self._add_statement(LogStatement(msg, var, time))

    def describe(self, fp=sys.stdout):
        init = []
        lines = []
        self._init_sequence.describe(init, init_section=True)
        self._sequence_stack[0].describe(lines)
        fp.write(f'Sequence:{self.name}\n')
        for line in init + lines:
            fp.write(line+'\n')
        fp.write('\n')

    def compile(self, generator, annotate=False):
        try:
            self._init_sequence.compile(generator, annotate)
            generator.start_main()
            self._sequence_stack[0].compile(generator, annotate)
            generator.end_main(self.end_time)
            self.modifies_frequency = generator.modifies_frequency
            self._compiled = True
        except Q1Exception as ex:
            logger.error(f'Compilation error on {self.name}', exc_info=True)
            msgs = [f'Error compiling {self.name}.']
            tb = []
            e = ex
            while e is not None:
                if isinstance(e, Q1SequenceError):
                    tb = e.traceback
                msgs.append(f'{type(e).__name__}: {e.args[0]}')
                e = e.__cause__
            q1_tb = '\n'.join(tb+msgs)
            self._dump_compile_state(generator, q1_tb, ex)
            # Use 'from None' to suppress original context
            # This avoids exposure of and confusion by q1pulse internals.
            raise Q1Exception(q1_tb) from None
        except Exception as ex:
            logger.error(f'Compilation error on {self.name}', exc_info=True)
            self._dump_compile_state(generator, None, ex)
            raise

    def _dump_compile_state(self, generator, q1_tb, exc):
        filename = os.path.join(os.getcwd(), '_q1pulse_dump.txt')
        print(f'**** Exception in {self.name} while compiling. ****')
        print(f'**** For details see: {filename} ****')
        with open(filename, 'w') as fp:
            fp.write(f'**** Exception in {self.name} while compiling. ****\n')
            if q1_tb:
                fp.write(q1_tb)
            fp.write('\n\n=== Q1Pulse Sequence ===\n')
            self.describe(fp)
            fp.write('\n=== Q1ASM till error ===\n')
            lines = generator.q1asm_lines()
            for line in lines:
                fp.write(line+'\n')
            fp.write('/-'*40 + '\n')
            fp.write('\n\n=== Full Exception ===\n\n')
            traceback.print_exception(exc, exc, None, file=fp)

    @property
    def current_time(self):
        return self.sequence.timeline.current_time + self._local_time

    def set_pulse_end(self, value):
        if self._local_time_active:
            self._local_time += max(0, value - self.current_time)
        else:
            self.sequence.timeline.set_pulse_end(value)

    @contextmanager
    def _local_timeline(self, duration=0, t_offset=0):
        if self._local_time_active:
            raise Q1InternalError('Local timeline already active')
        end_time = self.current_time + t_offset
        end_time += duration
        self._local_time = t_offset
        self._local_time_active = True
        yield
        self._local_time = 0
        self._local_time_active = False
        self.set_pulse_end(end_time)

    @property
    def end_time(self):
        return self.sequence.timeline.end_time

    def enter_loop(self, loop):
        loop_sequence = Sequence(self._timeline)
        if isinstance(loop, (RangeLoop, LinspaceLoop)):
            loop_statement = LoopStatement(self.end_time, loop_sequence, loop)
        elif isinstance(loop, ArrayLoop):
            loop_statement = ArrayLoopStatement(self.end_time, loop_sequence, loop)
        else:
            raise Q1InternalError('Unknown loop')
        self._add_statement(loop_statement)
        self._sequence_push(loop_sequence)

    def exit_loop(self, loop):
        if isinstance(loop, (RangeLoop, LinspaceLoop)):
            loop_end_statement = EndLoopStatement(self.end_time, loop)
        elif isinstance(loop, ArrayLoop):
            loop_end_statement = EndArrayLoopStatement(self.end_time, loop)
        else:
            raise Q1InternalError('Unknown loop')
        self._add_statement(loop_end_statement)
        self._sequence_pop()

    @contextmanager
    def _seq_repeat(self, n):
        ''' repeat loop not synchronizing with other sequencers. '''
        if n == 0:
            raise ValueError('n must be > 0')
        if n == 1:
            yield
        else:
            t_start = self.current_time
            loop = Loop(self._local_loop_cnt, n, local=True)
            self._local_loop_cnt += 1
            loop_sequence = Sequence(self._timeline)
            loop_statement = LoopStatement(self.current_time, loop_sequence, loop)
            self._add_statement(loop_statement)
            self._sequence_push(loop_sequence)

            yield

            loop_end_statement = EndLoopStatement(self.current_time, loop)
            self._add_statement(loop_end_statement)
            t_loop = self.current_time - t_start
            self._add_statement(LoopDurationStatement(n, t_loop))
            self._sequence_pop()
            self.set_pulse_end(t_start + n * t_loop)

    def _add_reg_wait(self, reg):
        self._add_statement(WaitRegStatement(self.end_time, reg))

    @property
    def trigger_counters(self):
        return self._trigger_counters

    def _add_trigger_counter(self, counter):
        self._trigger_counters.append(counter)

    def add_trigger_counter(self, trigger, threshold=1, invert=False):
        counter = TriggerCounter(trigger, threshold=threshold, invert=invert)
        self._add_trigger_counter(counter)
        return counter

    def latch_enable(self, enable, t_offset=0):
        if enable not in [0, 1, True, False]:
            raise ValueError('Valid values for enable are 0, 1, True, False')
        time = self.current_time + t_offset
        self._add_statement(LatchEnableStatement(time, enable))
        self.set_pulse_end(time)

    def latch_reset(self, t_offset=0):
        time = self.current_time + t_offset
        self._add_statement(LatchResetStatement(time))
        self.set_pulse_end(time)

    @contextmanager
    def conditional(self, counters, t_offset=0, evaluation_time=0):
        self.enter_conditional(counters, t_offset=t_offset, evaluation_time=evaluation_time)
        flags = CounterFlags(self)
        yield flags
        self.exit_conditional()

    def enter_conditional(self, counters, t_offset=0, evaluation_time=0):
        if self._conditional_block:
            raise Q1SyntaxError('Nested conditional is not supported')
        for counter in counters:
            if counter not in self._trigger_counters:
                raise Q1SyntaxError('Trigger counter {counter} not registered on sequencer')
        block = ConditionalBlockStatement(self.current_time+t_offset, counters)
        self._add_statement(block)
        self.set_pulse_end(self.current_time + t_offset + evaluation_time)
        self._conditional_block = block

    def exit_conditional(self):
        timeline = copy(self._timeline)
        self._conditional_block.close(timeline)
        self.set_pulse_end(self._conditional_block.end_time)
        self._conditional_block = None

    @contextmanager
    def condition(self, operator):
        self.enter_condition(operator)
        yield
        self.exit_condition()

    def enter_condition(self, operator):
        if self._conditional_block is None:
            raise Q1SyntaxError('Conditions must be wrapped in `conditional` section.')
        if self._in_condition:
            raise Q1SyntaxError('Conditions cannot be nested')
        self._in_condition = True
        timeline = copy(self._timeline)
        branch = BranchSequence(timeline, operator)
        self._conditional_block.add_branch(branch)
        self._sequence_push(branch)

    def exit_condition(self, end_time=None):
        if end_time is None:
            end_time = self.end_time
        self.add_comment(f'Condition end time: {end_time}')
        self._conditional_block.set_end_time(end_time)
        self._in_condition = False
        self._sequence_pop()
        self._last_timed_statement = self._conditional_block




class Q1Exception(Exception):
    '''
    Base for Q1Pulse exceptions.
    '''
    pass


class Q1InternalError(Q1Exception):
    '''
    Raised when something unexpected fails in Q1Pulse
    most likely caused by coding error in Q1Pulse.
    '''


class Q1NameError(Q1Exception):
    '''
    Raised when a register is not defined, or already defined.
    '''
    pass


class Q1TypeError(Q1Exception):
    '''
    Raised when expression or argument types do not match.
    '''


class Q1ValueError(Q1Exception):
    '''
    Raised when invalid value is passed.
    '''


class Q1SyntaxError(Q1Exception):
    '''
    Raised when invalid value is passed.
    '''


class Q1CompileError(Q1Exception):
    '''
    Raised when the error is detected during compilation.
    '''


class Q1SequenceError(Q1Exception):
    '''
    Raised when the error is detected during compilation.
    '''

    def __init__(self, msg, traceback):
        self.msg = msg
        self.traceback = traceback


class Q1MemoryError(Q1Exception):
    '''
    Raised when no free register is left to allocate.
    '''


class Q1TimingError(Q1Exception):
    '''
    Raised when the timing of commands doesn't fit.
    '''


class Q1StateError(Q1Exception):
    '''
    Raised when program or sequence is modified after compilation.
    '''


class Q1InputOverloaded(Q1Exception):
    '''
    Raised when the input of QRM is overloaded during acquisition.
    Exception can be suppressed with `q1pulse.set_exception_on_overload(False)`.
    '''
